img.padimg {
    padding-bottom: 5px;
    cursor: pointer;
}

li.hide {
  display: none !important;
}

li.show {
  display: list-item !important;
}

div.sphx-glr-clear {
    clear: left;
}

/* hide the top link */
div.sphx-glr-download-link-note {
  height: 0px;
  visibility: hidden;
}

/* Remove big whitespace to solve #1642 */
pre, div[class*="highlight-"] {
  clear: left;
}

/* Shrink overlarge warning box from PR #2169 */
div.admonition::after{content:none;}

/* override basic.css
 * xref https://github.com/sphinx-doc/sphinx/issues/10918 */
table.align-default {
    margin-left: 0px;
}
