//                                               -*- C++ -*-
/**
 *  @brief The test file of class LinearEnumerateFunction for standard methods
 *
 *  Copyright 2005-2026 Airbus-EDF-IMACS-ONERA-Phimeca
 *
 *  This library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "openturns/OT.hxx"
#include "openturns/OTtestcode.hxx"

using namespace OT;
using namespace OT::Test;

int main(int, char *[])
{
  TESTPREAMBLE;
  OStream fullprint(std::cout);

  try
  {
    UnsignedInteger size = 10;
    for (UnsignedInteger dimension = 1; dimension < 4; ++dimension)
    {
      LinearEnumerateFunction f(dimension);
      fullprint << "First " << size << " values for dimension " << dimension << std::endl;
      for (UnsignedInteger index = 0; index < size; ++index)
      {
        Indices multiIndex(f(index));
        fullprint << "index=" << index << ", multi-index=" << multiIndex << ", linear index=" << f.inverse(multiIndex) << std::endl;
      }
    }
    // Test getMarginal()
    fullprint << "Test getMarginal()" << std::endl;
    LinearEnumerateFunction enumerateFunction(10);
    Indices indices({0, 2, 4, 6, 9});
    EnumerateFunction marginalEnumerate(enumerateFunction.getMarginal(indices));
    assert_equal(marginalEnumerate.getDimension(), indices.getSize());
    for (UnsignedInteger index = 0; index < size; ++index)
    {
      Indices multiIndex(marginalEnumerate(index));
      fullprint << "index=" << index << ", multi-index=" << multiIndex << std::endl;
    }

  }
  catch (TestFailed & ex)
  {
    std::cerr << ex << std::endl;
    return ExitCode::Error;
  }

  return ExitCode::Success;
}
