//                                               -*- C++ -*-
/**
 *  @brief The external header file of OpenTURNS for subdir Experiments
 *
 *  Copyright 2005-2026 Airbus-EDF-IMACS-ONERA-Phimeca
 *
 *  This library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef OPENTURNS_OTEXPERIMENTS_HXX
#define OPENTURNS_OTEXPERIMENTS_HXX

#include "openturns/Axial.hxx"
#include "openturns/Box.hxx"
#include "openturns/Composite.hxx"
#include "openturns/Combinations.hxx"
#include "openturns/CombinatorialGenerator.hxx"
#include "openturns/CombinatorialGeneratorImplementation.hxx"
#include "openturns/Experiment.hxx"
#include "openturns/ExperimentImplementation.hxx"
#include "openturns/Factorial.hxx"
#include "openturns/KPermutations.hxx"
#include "openturns/SphereUniformNorm.hxx"
#include "openturns/SphereUniformNormImplementation.hxx"
#include "openturns/StratifiedExperiment.hxx"
#include "openturns/Tuples.hxx"

#endif /* OPENTURNS_OTEXPERIMENTS_HXX */
