import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


# Generate sample with the given plane
distribution = ot.JointDistribution([ot.Uniform(0, 1)] * 2)
size = 200
experiment = ot.LowDiscrepancyExperiment(ot.SobolSequence(), distribution, size)

sample = experiment.generate()

# Create an empty graph
graph = ot.Graph("Low Discrepancy experiment", "x1", "x2")

# Create the cloud
cloud = ot.Cloud(sample, "blue", "plus", "")

# Then, draw it
graph.add(cloud)

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)
