// the swig runtime header swigpyrun.h has no include guards
// https://github.com/swig/swig/pull/3325

#ifndef OPENTURNS_SWIGPYRUN_ONCE_H
#define OPENTURNS_SWIGPYRUN_ONCE_H
#include "openturns/swigpyrun.h"
#endif
