//                                               -*- C++ -*-
/**
 *  @brief The external header file of OpenTURNS for subdir Analytical
 *
 *  Copyright 2005-2026 Airbus-EDF-IMACS-ONERA-Phimeca
 *
 *  This library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this library.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef OPENTURNS_OTANALYTICAL_HXX
#define OPENTURNS_OTANALYTICAL_HXX

#include "openturns/Analytical.hxx"
#include "openturns/AnalyticalResult.hxx"
#include "openturns/StrongMaximumTest.hxx"
#include "openturns/FORM.hxx"
#include "openturns/FORMResult.hxx"
#include "openturns/MultiFORMResult.hxx"
#include "openturns/SORM.hxx"
#include "openturns/SORMResult.hxx"
#include "openturns/SystemFORM.hxx"
#include "openturns/MultiFORM.hxx"

#endif /* OPENTURNS_OTANALYTICAL_HXX */
