import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Create a Box Cox transformation
lambdas = [0.0, 0.1, 1.0, 1.5]

graph = ot.Graph("Box Cox transformations", "x", "y")
for i in range(len(lambdas)):
    boxCoxT = ot.BoxCoxTransform(lambdas[i])
    graph.add(boxCoxT.draw(0.1, 2.1))

graph.setLegends(["lambda = " + str(lam) for lam in lambdas])
graph.setLegendPosition("lower right")

fig = plt.figure(figsize=(8, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)
